/* (C) 2022 YiRing, Inc. */
package com.yiring.app.design.strategy.rule;

import com.yiring.app.constant.alarm.RelevanceParamEnum;
import com.yiring.app.design.strategy.IParamInitStrategy;
import com.yiring.app.design.strategy.IScanAlarmStrategy;
import com.yiring.app.domain.alarm.AlarmType;
import com.yiring.app.domain.location.LocationFence;
import com.yiring.app.domain.location.LocationFenceAlarm;
import com.yiring.app.domain.location.LocationTag;
import com.yiring.app.domain.location.LocationTagRepository;
import com.yiring.app.param.location.fence.LocationFenceJobParam;
import com.yiring.app.param.location.rule.LocationFenceRuleAddParam;
import com.yiring.app.param.location.rule.RuleParam;
import com.yiring.app.vo.location.fence.LocationFenceJobVo;
import com.yiring.common.core.Result;
import com.yiring.common.core.Status;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import javax.annotation.Resource;
import org.locationtech.jts.geom.Geometry;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

/**
 * 触发报警距离相关策略
 * @author tml
 * @version 1.0
 * @date 2022/4/29 9:33
 */
@Component
public class TriggerAlarmDistanceStrategy implements IParamInitStrategy<LocationFenceRuleAddParam>, IScanAlarmStrategy {

    @Resource
    private LocationTagRepository locationTagRepository;

    public TriggerAlarmDistanceStrategy() {
        Integer relevanceParam = RelevanceParamEnum.TRIGGER_ALARM_DISTANCE.getCode();
        LocationFenceRuleContext.register(relevanceParam, this);
    }

    @Override
    public Result<String> paramInit(LocationFenceRuleAddParam param) {
        RuleParam ruleParam = param.getRuleParam();
        if (ruleParam.getDistance() == null || ruleParam.getDistance() < 0) {
            return Result.no(Status.EXPECTATION_FAILED, "触发报警距离异常");
        }
        ruleParam.setDuration(-1);
        ruleParam.setTimeAndUser(new ArrayList<>());
        ruleParam.setTimeAndNumber(new ArrayList<>());
        return Result.ok();
    }

    @Override
    public LocationFenceJobVo scanAlarm(LocationFenceJobParam param) {
        RuleParam rule = param.getRule();
        Set<LocationTag> tagSet = param.getTagSet();
        Long mapId = param.getMapId();
        List<LocationFenceAlarm> oldAlarms = param.getOldAlarms();
        Long fenceId = param.getFenceId();
        Long alarmTypeId = param.getAlarmTypeId();
        Geometry geometry = param.getGeometry();
        LocationFenceJobVo jobVo = new LocationFenceJobVo();

        //查询出距离围栏规则之类距离的所有标签，封装成报警信息
        List<LocationTag> locationTags = locationTagRepository.findByDistance(geometry, rule.getDistance());
        List<LocationFenceAlarm> newAlarms = locationTags
            .stream()
            .map(e -> {
                LocationFence fence = LocationFence.builder().id(fenceId).build();
                AlarmType alarmType = AlarmType.builder().id(alarmTypeId).build();
                return LocationFenceAlarm
                    .builder()
                    .fence(fence)
                    .point(e.getPoint())
                    .areaId(mapId)
                    .user(e.getUser())
                    .tag(e)
                    .startTime(LocalDateTime.now())
                    .type(alarmType)
                    .status(LocationFenceAlarm.Status.ING)
                    .build();
            })
            .collect(Collectors.toList());
        //取两个集合没有相交的部分，newAlarms不重合的部分代表需要添加的报警，oldAlarms不重合的部分代表需要结束的报警
        if (!CollectionUtils.isEmpty(oldAlarms)) {
            Set<Long> newSet = newAlarms.stream().map(e -> e.getTag().getId()).collect(Collectors.toSet());
            Set<Long> oldSet = oldAlarms.stream().map(e -> e.getTag().getId()).collect(Collectors.toSet());
            newAlarms =
                newAlarms.stream().filter(e -> !oldSet.contains(e.getTag().getId())).collect(Collectors.toList());
            oldAlarms =
                oldAlarms.stream().filter(e -> !newSet.contains(e.getTag().getId())).collect(Collectors.toList());
        }
        jobVo.setAddAlarm(newAlarms);
        jobVo.setModifyAlarm(oldAlarms);
        return jobVo;
    }
}
