/* (C) 2022 YiRing, Inc. */
package com.yiring.app.service.message.impl;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.yiring.app.domain.district.District;
import com.yiring.app.domain.district.DistrictRepository;
import com.yiring.app.domain.key.*;
import com.yiring.app.domain.location.*;
import com.yiring.app.domain.log.ZyRealtimeLog;
import com.yiring.app.domain.log.ZyRealtimeLogRepository;
import com.yiring.app.param.key.KeyAlarmAddParam;
import com.yiring.app.service.message.PositionMessageService;
import com.yiring.app.util.GeoUtils;
import com.yiring.app.vo.equipment.EquipmentVo;
import com.yiring.auth.domain.dept.Department;
import com.yiring.auth.domain.user.User;
import com.yiring.auth.domain.user.UserRepository;
import com.yiring.common.annotation.Times;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.annotation.Resource;
import javax.persistence.criteria.Expression;
import javax.persistence.criteria.Predicate;
import lombok.extern.slf4j.Slf4j;
import org.locationtech.jts.geom.Point;
import org.springframework.beans.BeanUtils;
import org.springframework.data.domain.*;
import org.springframework.messaging.simp.SimpMessagingTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * 真源消息处理实现类
 *
 * @author Jim
 * @version 0.1
 * 2022/5/9 10:22
 */

@Slf4j
@Service
@Transactional(rollbackFor = Exception.class)
public class PositionMessageServiceImpl implements PositionMessageService {

    @Resource
    LocationTagRepository locationTagRepository;

    @Resource
    LocationLogRepository locationLogRepository;

    @Resource
    LocationFenceRepository locationFenceRepository;

    @Resource
    DistrictRepository districtRepository;

    @Resource
    LocationTurnoverRepository locationTurnoverRepository;

    @Resource
    SimpMessagingTemplate simpMessagingTemplate;

    @Resource
    ZyRealtimeLogRepository zyRealtimeLogRepository;

    @Resource
    KeyAlarmRepository keyAlarmRepository;

    @Resource
    KeyAlarmLogRepository keyAlarmLogRepository;

    @Resource
    KeyAlarmRuleRepository keyAlarmRuleRepository;

    @Resource
    UserRepository userRepository;

    @Times("Message Consume")
    @Override
    public void consume(String message) {
        // 将消息转换成 JSON 格式
        JSONObject info = JSON.parseObject(message);
        log.info("Receiver Message: {}", info);

        // 解构消息内容
        JSONObject data = info.getJSONObject("params");
        String method = info.getString("method");

        // 记录日志
        ZyRealtimeLog realtimeLog = ZyRealtimeLog.builder().method(method).raw(info).build();
        zyRealtimeLogRepository.save(realtimeLog);

        // 业务处理
        switch (method) {
            // 实时定位
            case "position" -> processPositionMessage(data);
            // 设备低电量
            case "lowPower" -> processLowPowerMessage(data);
            // 设备状态变更
            case "deviceStatus" -> processDeviceStatusMessage(data);
            // 按键报警
            case "keyWarning" -> processKeyWarningMessage(data);
            // 围栏报警
            case "enclosure" -> log.warn("Ignore Message Type [enclosure]: {}", info);
            default -> log.warn("Unknown Message Type: {}", info);
        }
    }

    /**
     * 处理定位消息
     * @param data 消息内容
     */
    public void processPositionMessage(JSONObject data) {
        // TODO
        log.info("Position Message: {}", data);

        // 包装消息
        // TODO
        // 1. 解析消息内容，进行围栏、出入标识判断等处理，将定位记录录入数据库
        // 2. 创建一条需要进行消息推送的记录
        // 3. 将记录推送的消息推送模块
        // 4. 检查是否触发围栏告警，记录告警数据，并推送消息

        // 查询定位标签
        String tagId = data.getString("tagId");
        Example<LocationTag> example = Example.of(LocationTag.builder().code(tagId).build());
        Optional<LocationTag> optional = locationTagRepository.findOne(example);
        if (optional.isEmpty()) {
            log.warn("Tag Not Found: {}", tagId);
            return;
        }

        // 构建复合主键
        LocationTag tag = optional.get();
        TagTimeId id = new TagTimeId(tag);

        // 定位时间
        Instant instant = Instant.ofEpochMilli(data.getLongValue("locationTime"));
        LocalDateTime locationTime = LocalDateTime.ofInstant(instant, ZoneId.systemDefault());

        // 定位的基本信息
        LocationLog locationLog = LocationLog
            .builder()
            .id(id)
            .raw(data)
            .locationTime(locationTime)
            .areaId(data.getLong("areaId"))
            .floor(data.getString("floor"))
            .silent(data.getBoolean("silent"))
            .volt(data.getInteger("volt"))
            .voltUnit(data.getString("voltUnit"))
            .out(data.getBoolean("out"))
            .build();

        // 获取定位卡当前绑定的用户
        User user = tag.getUser();
        if (user != null) {
            // 定位标签当时所属的用户
            locationLog.setUser(user);
            // 查询当前用户的状态
            locationLog.setStatus(user.getStatus());
        }

        // 设置空间点位信息
        Point point = GeoUtils.createPoint(
            data.getDoubleValue("longitude"),
            data.getDoubleValue("latitude"),
            data.getDoubleValue("altitude")
        );
        locationLog.setPoint(point);

        // 定位信标
        Set<String> beaconCodes = Arrays
            .stream(data.getString("beacons").split(","))
            .map(beacon -> beacon.replaceAll("\\(.*\\)", ""))
            .collect(Collectors.toSet());
        locationLog.setBeacons(new JSONArray().fluentAddAll(beaconCodes));

        // 计算出入标记（围栏、区域）
        List<LocationTurnover> turnovers = new ArrayList<>();

        // 查询定位在围栏内的围栏信息（同时根据地图区域查询，用来区分不同楼层的围栏）
        List<LocationFence> fences = locationFenceRepository.findByGeometryContains(locationLog.getAreaId(), point);
        Set<Long> fenceIds = fences.stream().map(LocationFence::getId).collect(Collectors.toSet());
        locationLog.setFences(new JSONArray().fluentAddAll(fenceIds));
        // 计算围栏进出
        for (LocationFence fence : fences) {
            // 查询当前围栏的防抖时间内的所有定位记录
            List<LocationLog> logs = findByTagTimeIdAndThreshold(id, fence.getThreshold());
            List<JSONArray> list = logs.stream().map(LocationLog::getFences).toList();
            if (list.isEmpty()) {
                continue;
            }

            // 检查是否进入区域
            Boolean isEnter = checkEnter(list, fence.getId());
            if (isEnter != null) {
                // 检查是否为重复进入/退出
                Optional<LocationTurnover> enter = findRepeatTurnoverRecord(
                    fence.getId(),
                    LocationTurnover.Type.FENCE,
                    id.getTag(),
                    isEnter
                );
                if (enter.isEmpty()) {
                    // 尝试将上一次记录标记为非最新记录
                    trySetPrevTurnoverExpired(fence.getId(), LocationTurnover.Type.FENCE, id.getTag());

                    // 当前标签进入/离开围栏的动作标记为最新记录
                    LocationTurnover turnover = LocationTurnover
                        .builder()
                        .enter(isEnter)
                        .type(LocationTurnover.Type.FENCE)
                        .sourceId(fence.getId())
                        .tag(id.getTag())
                        .time(id.getTime())
                        .user(id.getTag().getUser())
                        .isLatest(true)
                        .build();
                    turnovers.add(turnover);

                    // 更新围栏内的标签
                    fence.setTags(resetInTags(tag, isEnter, fence.getTags()));
                    // 有人员进出围栏，需要检查是否触发围栏报警规则
                    // TODO: 通过定时任务调度异步实现，提高定位消息消费能力
                    // 1. 判断是否触发围栏报警规则，触发则记录报警记录，同时记录报警记录触发所需推送的消息
                    // 2. 同时进行 WebSocket 消息推送
                }
            }
        }

        // 查询定位在区域内的区域信息
        List<District> districts = districtRepository.findByGeometryContains(point);
        Set<Long> districtIds = districts.stream().map(District::getId).collect(Collectors.toSet());
        locationLog.setDistricts(new JSONArray().fluentAddAll(districtIds));
        // 计算区域进出
        for (District district : districts) {
            // 查询当前区域的防抖时间内的所有定位记录
            List<LocationLog> logs = findByTagTimeIdAndThreshold(id, district.getDebouncingDuration());
            List<JSONArray> list = logs.stream().map(LocationLog::getDistricts).toList();
            if (list.isEmpty()) {
                continue;
            }

            // 检查是否进入区域
            Boolean isEnter = checkEnter(list, district.getId());
            if (isEnter != null) {
                // 检查是否为重复进入/退出
                Optional<LocationTurnover> enter = findRepeatTurnoverRecord(
                    district.getId(),
                    LocationTurnover.Type.DISTRICT,
                    id.getTag(),
                    isEnter
                );
                if (enter.isEmpty()) {
                    // 尝试将上一次记录标记为非最新记录
                    trySetPrevTurnoverExpired(district.getId(), LocationTurnover.Type.DISTRICT, id.getTag());

                    // 当前标签进入/离开围栏的动作标记为最新记录
                    LocationTurnover turnover = LocationTurnover
                        .builder()
                        .enter(isEnter)
                        .type(LocationTurnover.Type.DISTRICT)
                        .sourceId(district.getId())
                        .tag(id.getTag())
                        .time(id.getTime())
                        .user(id.getTag().getUser())
                        .isLatest(true)
                        .build();
                    turnovers.add(turnover);

                    // 更新区域内的标签
                    district.setTags(resetInTags(tag, isEnter, district.getTags()));
                }
            }
        }

        // 写入定位数据
        locationLogRepository.saveAndFlush(locationLog);

        // 更新定位标签卡状态信息
        tag.setOut(locationLog.getOut());
        tag.setPoint(locationLog.getPoint());
        tag.setVolt(locationLog.getVolt());
        tag.setVoltUnit(locationLog.getVoltUnit());
        tag.setSilent(locationLog.getSilent());
        locationTagRepository.save(tag);

        // 更新围栏内的标签数据
        locationFenceRepository.saveAll(fences);
        // 更新区域内的标签数据
        districtRepository.saveAll(districts);

        // 写入围栏/区域进出记录
        locationTurnoverRepository.saveAll(turnovers);

        // WebSocket 推送定位消息
        // 消息内容需要确定 TODO
        JSONObject message = new JSONObject();
        message.put("type", "location");
        message.put("time", id.getTime());
        message.put("tagId", id.getTag().getId());
        message.put("tagCode", id.getTag().getCode());
        message.put("entityId", locationLog.getPoint());
        message.put("point", locationLog.getPoint());
        simpMessagingTemplate.convertAndSend("/topic/position", message.toJSONString());
    }

    /**
     * 根据定位日志 ID 和防抖时间查询定位日志
     * @param id 定位日志 ID
     * @param threshold 防抖时间
     * @return 定位日志集合
     */
    public List<LocationLog> findByTagTimeIdAndThreshold(TagTimeId id, int threshold) {
        return locationLogRepository.findAll((root, query, cb) -> {
            Predicate predicate = cb.conjunction();
            List<Expression<Boolean>> expressions = predicate.getExpressions();
            expressions.add(cb.equal(root.get("id").get("tag"), id.getTag()));
            expressions.add(cb.greaterThanOrEqualTo(root.get("id").get("time"), id.getTime().minusSeconds(threshold)));
            expressions.add(cb.lessThanOrEqualTo(root.get("id").get("time"), id.getTime()));
            return predicate;
        });
    }

    /**
     * 根据防抖期间进入的区域/围栏以及新产生的围栏/区域记录，判断是否进入
     * @param array 区域/围栏集合
     * @param id 区域/围栏 ID
     * @return true 进入，false 退出，null 未发生变化
     */
    public Boolean checkEnter(List<JSONArray> array, Long id) {
        long count = array.stream().filter(ids -> ids != null && ids.contains(id)).count();
        Boolean isEnter = null;
        if (count == array.size()) {
            isEnter = true;
        } else if (count == 0) {
            isEnter = false;
        }
        return isEnter;
    }

    /**
     * 查找
     * @param sourceId 区域/围栏 ID
     * @param type 类型（区域/围栏）
     * @param tag 定位标签
     * @param enter true 进入，false 退出
     * @return true 重复进入，false 非重复进入
     */
    public Optional<LocationTurnover> findRepeatTurnoverRecord(
        Long sourceId,
        LocationTurnover.Type type,
        LocationTag tag,
        Boolean enter
    ) {
        Pageable pageable = PageRequest.of(0, 1, Sort.by(Sort.Order.desc(LocationTurnover.Fields.time)));
        Page<LocationTurnover> page = locationTurnoverRepository.findAll(
            (root, query, cb) -> {
                Predicate predicate = cb.conjunction();
                List<Expression<Boolean>> expressions = predicate.getExpressions();
                expressions.add(cb.equal(root.get(LocationTurnover.Fields.sourceId), sourceId));
                expressions.add(cb.equal(root.get(LocationTurnover.Fields.type), type));
                expressions.add(cb.equal(root.get(LocationTurnover.Fields.tag), tag));
                return predicate;
            },
            pageable
        );

        if (page.getTotalElements() > 0) {
            // 检查是否重复进入
            Stream<LocationTurnover> stream = page.get();
            if (enter == null || stream.anyMatch(turnover -> turnover.getEnter() == enter)) {
                return stream.findFirst();
            }
        }

        return Optional.empty();
    }

    /**
     * 尝试设置上一个进出记录为过期标识
     * @param sourceId 区域/围栏 ID
     * @param type 类型（区域/围栏）
     * @param tag 定位标签
     */
    public void trySetPrevTurnoverExpired(Long sourceId, LocationTurnover.Type type, LocationTag tag) {
        Optional<LocationTurnover> record = findRepeatTurnoverRecord(sourceId, type, tag, null);
        if (record.isPresent()) {
            LocationTurnover turnoverRecord = record.get();
            turnoverRecord.setIsLatest(false);
            locationTurnoverRepository.save(turnoverRecord);
        }
    }

    public Set<LocationTag> resetInTags(LocationTag tag, Boolean enter, Set<LocationTag> tags) {
        if (Boolean.TRUE.equals(enter)) {
            tags.add(tag);
        } else {
            tags.remove(tag);
        }
        return tags;
    }

    /**
     * 处理低电量报警消息
     * @param data 消息内容
     */
    public void processLowPowerMessage(JSONObject data) {
        // TODO
        log.info("LowPower Message: {}", data);
    }

    /**
     * 处理设备状态更新消息
     * @param data 消息内容
     */
    public void processDeviceStatusMessage(JSONObject data) {
        // TODO
        log.info("DeviceStatus Message: {}", data);
        EquipmentVo equipmentVo = data.toJavaObject(EquipmentVo.class);
        if (equipmentVo.getDeviceType().equals("BTI")) {
            Optional<LocationTag> one = locationTagRepository.findOne(
                Example.of(LocationTag.builder().code(equipmentVo.getDeviceId()).build())
            );
            if (one.isEmpty()) {
                log.error("没找到该标签");
                return;
            }
            //修改标签电量
            LocationTag locationTag = LocationTag
                .builder()
                .code(equipmentVo.getDeviceId())
                .volt(equipmentVo.getVolt())
                .voltUnit(equipmentVo.getField_21())
                .build();
            BeanUtils.copyProperties(one.get(), locationTag);
            locationTagRepository.saveAndFlush(locationTag);
        }
    }

    /**
     * 处理按键报警消息
     * @param data 消息内容
     */
    public void processKeyWarningMessage(JSONObject data) {
        // TODO
        log.info("KeyWarning Message: {}", data);
        KeyAlarmAddParam keyAlarmAddParam = data.toJavaObject(KeyAlarmAddParam.class);
        KeyAlarm keyAlarm = KeyAlarmAddParam.transform(keyAlarmAddParam);
        keyAlarm.setPoint(
            GeoUtils.xyzToPoint(
                keyAlarm.getX().doubleValue(),
                keyAlarm.getY().doubleValue(),
                keyAlarm.getZ().doubleValue()
            )
        );

        Optional<User> userRepositoryOne = userRepository.findOne(
            Example.of(User.builder().linkId(keyAlarmAddParam.getEntityId()).build())
        );

        if (userRepositoryOne.isEmpty()) {
            //            throw new RuntimeException("没找到该用户");
            log.warn("Tag Not Found: {}", "没找到该用户");
            return;
        }

        User user = userRepositoryOne.get();

        keyAlarm.setLeader(user);
        keyAlarm.setEnable(true);
        Optional<KeyAlarmRule> keyAlarmRuleOptional = keyAlarmRuleRepository.findOne(
            Example.of(
                KeyAlarmRule.builder().department(Department.builder().id(user.getDepartment().getId()).build()).build()
            )
        );

        if (keyAlarmRuleOptional.isEmpty()) {
            //            throw new RuntimeException("该部门未设置报警规则");
            log.warn("Tag Not Found: {}", "该部门未设置报警规则");
            return;
        }

        Optional<LocationTag> locationTagOptional = locationTagRepository.findOne(
            Example.of(LocationTag.builder().code(keyAlarmAddParam.getTagId()).build())
        );

        if (locationTagOptional.isEmpty()) {
            log.warn("Tag Not Found: {}", keyAlarmAddParam.getTagId());
            return;
        }
        keyAlarm.setCode(locationTagOptional.get().getCode());

        final KeyAlarm finalKeyAlarm = keyAlarmRepository.save(keyAlarm);

        KeyAlarmRule keyAlarmRule = keyAlarmRuleOptional.get();
        Set<User> users = keyAlarmRule.getUsers();

        HashSet<KeyAlarmLog> keyAlarmLogs = new HashSet<>();
        users.forEach(leader -> {
            String[] split = keyAlarmRule.getTypes().split(",");
            for (String type : split) {
                keyAlarmLogs.add(
                    KeyAlarmLog.builder().status(false).user(leader).types(type).keyAlarm(finalKeyAlarm).build()
                );
            }
        });

        keyAlarmLogRepository.saveAll(keyAlarmLogs);
    }
}
